/*
 * First.h
 *
 *  Created on: 9. 6. 2015
 *	  Author: Tomas Pecka, Jan Travnicek
 */

#ifndef FIRST_H_
#define FIRST_H_

#include <alib/vector>
#include <alib/set>
#include <alib/variant>

#include <string/Epsilon.h>
#include <grammar/Grammar.h>

namespace grammar {

namespace parsing {

class First {
	template < class SymbolType >
	static ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > first ( const ext::set < SymbolType > & terminals, const ext::set < SymbolType > & nonterminals, const ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > & firstOfNonterminal, const ext::vector < SymbolType > & rhs );

	template < class SymbolType >
	static ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > first ( const ext::set < SymbolType > & terminals, const ext::set < SymbolType > & nonterminals, const ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > & rules );

public:
	template < class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static ext::map < ext::vector < SymbolType >, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > first ( const T & grammar );

	template < class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > first ( const T & grammar, const ext::vector < SymbolType > & rhs );
};

template < class SymbolType >
ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > First::first ( const ext::set < SymbolType > & terminals, const ext::set < SymbolType > & nonterminals, const ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > & firstOfNonterminal, const ext::vector < SymbolType > & rhs ) {
	 // 1. FIRST(\varepsilon) = { \varepsilon }
	if ( rhs.size ( ) == 0 ) {
		return { string::Epsilon < SymbolType >::EPSILON };
	}

	 // 2. FIRST(a) = { a } forall a \in T
	else if ( terminals.count ( rhs[0] ) ) {
		return { rhs[0] };
	}

	 // 4. FIRST(A \alpha) = first(A) if A \in N and \varepsilon \notin first(A)
	else if ( nonterminals.count ( rhs[0] ) && !firstOfNonterminal.find ( rhs[0] )->second.count ( ext::variant < SymbolType, string::Epsilon < SymbolType > >::template from < string::Epsilon < SymbolType > > ( ) ) ) {
		return firstOfNonterminal.find ( rhs[0] )->second;
	}

	 // 5. FIRST(A \alpha) = (first(A) - \varepsilon) \cup FIRST(\alpha) if A \in N and \varepsilon \in first(A)
	else if ( nonterminals.count ( rhs[0] ) && firstOfNonterminal.find ( rhs[0] )->second.count ( ext::variant < SymbolType, string::Epsilon < SymbolType > >::template from < string::Epsilon < SymbolType > > ( ) ) ) {
		ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > res = firstOfNonterminal.find ( rhs[0] )->second;
		res.erase ( string::Epsilon < SymbolType >::EPSILON );

		ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > next = first ( terminals, nonterminals, firstOfNonterminal, ext::vector < SymbolType > ( rhs.begin ( ) + 1, rhs.end ( ) ) );
		res.insert ( next.begin ( ), next.end ( ) );
		return res;
	} else {
		throw exception::CommonException ( "Cant be reached" );
	}
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > First::first ( const ext::set < SymbolType > & terminals, const ext::set < SymbolType > & nonterminals, const ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > & rules ) {
	/*
	 *
	 * 1. foreach A \in N: first(A) = \emptyset
	 * 2. foreach A \rightarrow \alpha:
	 *	first(A) = first(A) \cup FIRST(\alpha)
	 * 3. repeat step 2 if at least one set first(A) has changed
	 *
	 */
	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > firstOfNonterminal1;

	for ( const SymbolType & nonterminal : nonterminals )
		firstOfNonterminal1[nonterminal];

	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > firstOfNonterminal2 = firstOfNonterminal1;

	do {
		for ( const std::pair < const SymbolType, ext::set < ext::vector < SymbolType > > > & rule : rules )
			for ( const ext::vector < SymbolType > & rhs : rule.second ) {
				ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > newFirst = first ( terminals, nonterminals, firstOfNonterminal1, rhs );
				firstOfNonterminal2[rule.first].insert ( newFirst.begin ( ), newFirst.end ( ) );
			}

		if ( firstOfNonterminal1 == firstOfNonterminal2 )
			break;

		firstOfNonterminal1 = std::move ( firstOfNonterminal2 );
		firstOfNonterminal2.clear ( );

	} while ( true );

	return firstOfNonterminal1;
}

template < class T, class SymbolType >
ext::map < ext::vector < SymbolType >, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > First::first ( const T & grammar ) {
	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > firstNt = first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), grammar.getRawRules ( ) );

	ext::map < ext::vector < SymbolType >, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > res;

	for ( const std::pair < const SymbolType, ext::set < ext::vector < SymbolType > > > & rule : grammar.getRawRules ( ) )
		for ( const ext::vector < SymbolType > & rhs : rule.second )
			res.insert ( std::make_pair ( rhs, first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), firstNt, rhs ) ) );

	return res;
}

template < class T, class SymbolType >
ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > First::first ( const T & grammar, const ext::vector < SymbolType > & rhs ) {
	ext::map < SymbolType, ext::set < ext::variant < SymbolType, string::Epsilon < SymbolType > > > > firstNt = first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), grammar.getRawRules ( ) );

	return first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), firstNt, rhs );
}

} /* namespace parsing */

} /* namespace grammar */

#endif /* FIRST_H_ */
