/*
 * NonlinearVariableSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef NONLINEAR_VARIABLE_SYMBOL_H_
#define NONLINEAR_VARIABLE_SYMBOL_H_

#include "Symbol.h"
#include "UniqueSymbol.h"

#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace alphabet {

/**
 * \brief
 * Represents the nonlinear variable symbol used in a nonlinear tree pattern.
 */
template < class SymbolType = DefaultSymbolType >
class NonlinearVariableSymbol : public SymbolBase {
	/**
	 * \brief the symbol of the nonlinear variable.
	 */
	SymbolType m_symbol;

public:
	/**
	 * \brief
	 * Creates a new instance of the nonlinear variable with some underlying base symbol.
	 */
	explicit NonlinearVariableSymbol ( SymbolType symbol );

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const &;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) &&;

	/**
	 * Getter of the nonlinear variable's symbol
	 *
	 * \return the symbol of the nonlinear variable
	 */
	const SymbolType & getSymbol ( ) const &;

	/**
	 * Getter of the nonlinear variable's symbol
	 *
	 * \return the symbol of the nonlinear variable
	 */
	SymbolType && getSymbol ( ) &&;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	virtual int compare ( const NonlinearVariableSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual SymbolBase * inc ( ) &&;
};

template < class SymbolType >
NonlinearVariableSymbol < SymbolType >::NonlinearVariableSymbol ( SymbolType symbol ) : m_symbol ( std::move ( symbol ) ) {
}

template < class SymbolType >
SymbolBase * NonlinearVariableSymbol < SymbolType >::clone ( ) const & {
	return new NonlinearVariableSymbol ( * this );
}

template < class SymbolType >
SymbolBase * NonlinearVariableSymbol < SymbolType >::clone ( ) && {
	return new NonlinearVariableSymbol ( std::move ( * this ) );
}

template < class SymbolType >
const SymbolType & NonlinearVariableSymbol < SymbolType >::getSymbol ( ) const & {
	return m_symbol;
}

template < class SymbolType >
SymbolType && NonlinearVariableSymbol < SymbolType >::getSymbol ( ) && {
	return std::move ( m_symbol );
}

template < class SymbolType >
int NonlinearVariableSymbol < SymbolType >::compare ( const NonlinearVariableSymbol & other ) const {
	static ext::compare < decltype ( m_symbol ) > comp;

	return comp ( m_symbol, other.m_symbol );
}

template < class SymbolType >
void NonlinearVariableSymbol < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(NonlinearVariableSymbol " << m_symbol << ")";
}

template < class SymbolType >
NonlinearVariableSymbol < SymbolType >::operator std::string ( ) const {
	return "$" + ext::to_string ( m_symbol );
}

template < class SymbolType >
SymbolBase * NonlinearVariableSymbol < SymbolType >::inc ( ) && {
	return new UniqueSymbol ( Symbol ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace alphabet */

namespace core {

template < class SymbolType >
struct normalize < alphabet::NonlinearVariableSymbol < SymbolType > > {
	static alphabet::NonlinearVariableSymbol < > eval ( alphabet::NonlinearVariableSymbol < SymbolType > && value ) {
		return alphabet::NonlinearVariableSymbol < > ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( value ).getSymbol ( ) ) );
	}
};

} /* namespace core */

#endif /* NONLINEAR_VARIABLE_SYMBOL_H_ */
