/*
 * RankedSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Apr 10, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_SYMBOL_H_
#define RANKED_SYMBOL_H_

#include "SymbolBase.h"
#include <common/ranked_symbol.hpp>

#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace alphabet {

/**
 * \brief
 * Represents the adaptor of ranked symbol to class in symbol hierarchy.
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType = DefaultSymbolType, class RankType = DefaultRankType >
class RankedSymbol : public SymbolBase, public common::ranked_symbol < SymbolType, RankType > {
public:
	/**
	 * \brief
	 * The constructors of raw ranked symbol are inherited here
	 */
	using common::ranked_symbol < SymbolType, RankType >::ranked_symbol;

	/**
	 * \brief
	 * Creates a new instance of the ranked symbol based on a raw ranked symbol
	 */
	explicit RankedSymbol ( common::ranked_symbol < SymbolType, RankType > symbol ) : common::ranked_symbol < SymbolType, RankType > ( std::move ( symbol ) ) {
	}

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const &;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) &&;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	virtual int compare ( const RankedSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator ==
	 */
	using SymbolBase::operator ==;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator !=
	 */
	using SymbolBase::operator !=;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator <
	 */
	using SymbolBase::operator <;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator <=
	 */
	using SymbolBase::operator <=;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator >
	 */
	using SymbolBase::operator >;

	/**
	 * @copydoc base::CommonBaseMiddle < ObjectBase >::operator >=
	 */
	using SymbolBase::operator >=;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual SymbolBase * inc ( ) &&;
};

template < class SymbolType, class RankType >
SymbolBase* RankedSymbol < SymbolType, RankType >::clone ( ) const & {
	return new RankedSymbol(*this);
}

template < class SymbolType, class RankType >
SymbolBase* RankedSymbol < SymbolType, RankType >::clone() && {
	return new RankedSymbol(std::move(*this));
}

template < class SymbolType, class RankType >
int RankedSymbol < SymbolType, RankType >::compare(const RankedSymbol& other) const {
	return static_cast < const common::ranked_symbol < SymbolType, RankType > & > ( * this ).compare ( static_cast < const common::ranked_symbol < SymbolType, RankType > & > ( other ) );
}

template < class SymbolType, class RankType >
void RankedSymbol < SymbolType, RankType >::operator>>(std::ostream& out) const {
	out << static_cast < const common::ranked_symbol < SymbolType, RankType > & > ( * this );
}

template < class SymbolType, class RankType >
RankedSymbol < SymbolType, RankType >::operator std::string () const {
	return (std::string) static_cast < const common::ranked_symbol < SymbolType, RankType > & > ( * this );
}

template < class SymbolType, class RankType >
SymbolBase* RankedSymbol < SymbolType, RankType >::inc() && {
	this->m_symbol.inc();
	return NULL;
}

} /* namespace alphabet */

namespace core {

template < class SymbolType, class RankType >
struct normalize < common::ranked_symbol < SymbolType, RankType > > {
	static common::ranked_symbol < > eval ( common::ranked_symbol < SymbolType, RankType > && value ) {
		return alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( value ) );
	}
};

template < class SymbolType, class RankType >
struct normalize < alphabet::RankedSymbol < SymbolType, RankType > > {
	static alphabet::RankedSymbol < > eval ( alphabet::RankedSymbol < SymbolType, RankType > && value ) {
		return alphabet::RankedSymbol < > ( normalize < common::ranked_symbol < SymbolType, RankType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* RANKED_SYMBOL_H_ */
