/*
 * StartSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Jun 19, 2014
 *      Author: Jan Travnicek
 */

#ifndef START_SYMBOL_H_
#define START_SYMBOL_H_

#include "SymbolBase.h"
#include "Symbol.h"
#include <object/Object.h>
#include <common/ranked_symbol.hpp>

namespace alphabet {

/**
 * \brief
 * Represents start symbol used as initializing symbol of a string.
 */
class StartSymbol : public SymbolBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the symbol.
	 */
	explicit StartSymbol ( );

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const &;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) &&;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	virtual int compare ( const StartSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const;

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if <   std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if < ! std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual SymbolBase * inc ( ) &&;
};

template < typename Base >
inline typename std::enable_if < std::is_integral < Base >::value, Base >::type StartSymbol::instance ( ) {
	return '^';
}

template < >
inline object::Object StartSymbol::instance < object::Object > ( ) {
	return object::Object ( StartSymbol ( ) );
}

template < >
inline std::string StartSymbol::instance < std::string > ( ) {
	return std::string ( 1, StartSymbol::instance < char > ( ) );
}

template < >
inline StartSymbol StartSymbol::instance < StartSymbol > ( ) {
	return StartSymbol ( );
}

template < >
inline Symbol StartSymbol::instance < Symbol > ( ) {
	return Symbol ( StartSymbol ( ) );
}

template < >
inline common::ranked_symbol < > StartSymbol::instance < common::ranked_symbol < > > ( ) {
	return common::ranked_symbol < > ( StartSymbol::instance < DefaultSymbolType > ( ), DefaultRankType ( 0 ) );
}

} /* namespace alphabet */

#endif /* START_SYMBOL_H_ */
