/*
 * UniqueSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Mar 26, 2013
 *      Author: Jan Travnicek
 */

#ifndef SYMBOL_ID_SYMBOL_H_
#define SYMBOL_ID_SYMBOL_H_

#include "Symbol.h"
#include <primitive/Integer.h>

namespace alphabet {

/**
 * \brief
 * Represents symbol based on other symbol with additional differentiation counter.
 *
 * The UniqueSymbol is used in situations when some symbol is given and one would like to create new based on it but different. For example q vs. q' vs q'', etc.
 *
 * Primarly the class is used from inc method.
 */
class UniqueSymbol : public SymbolBase {
	/**
	 * The underlying symbol.
	 */
	Symbol m_symbol;

	/**
	 * The differentiating number.
	 */
	primitive::Integer m_id;

public:
	/**
	 * \brief
	 * Creates new symbol with given name and set differentiating number.
	 *
	 * \param symbol the underlying symbol
	 * \param id the differentiating number
	 */
	explicit UniqueSymbol ( Symbol symbol, primitive::Integer id );

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const &;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) &&;

	/**
	 * Getter of the underlying symbol.
	 *
	 * \return the underlying symbol
	 */
	const Symbol & getSymbol ( ) const &;

	/**
	 * Getter of the underlying symbol.
	 *
	 * \return the underlying symbol
	 */
	Symbol && getSymbol ( ) &&;

	/**
	 * Getter of the differentiating number.
	 *
	 * \return the differentiating number
	 */
	const primitive::Integer & getId ( ) const &;

	/**
	 * Getter of the differentiating number.
	 *
	 * \return the differentiating number
	 */
	primitive::Integer && getId ( ) &&;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	virtual int compare ( const UniqueSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual SymbolBase * inc ( ) &&;
};

} /* namespace alphabet */

#endif /* SYMBOL_ID_SYMBOL_H_ */
