/*
 * WildcardSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef SUBTREE_WILDCARD_SYMBOL_H_
#define SUBTREE_WILDCARD_SYMBOL_H_

#include "SymbolBase.h"
#include "Symbol.h"
#include "RankedSymbol.h"
#include <object/Object.h>

namespace alphabet {

/**
 * \brief
 * Represents a wildcard used as representation of anything here in a string or a tree.
 */
class WildcardSymbol : public SymbolBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the symbol.
	 */
	explicit WildcardSymbol ( );

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const &;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) &&;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	virtual int compare ( const WildcardSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const;

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if <   std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if < ! std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual SymbolBase * inc ( ) &&;
};

template < typename Base >
inline typename std::enable_if < std::is_integral < Base >::value, Base >::type WildcardSymbol::instance ( ) {
	return '*';
}

template < >
inline object::Object WildcardSymbol::instance < object::Object > ( ) {
	return object::Object ( WildcardSymbol ( ) );
}

template < >
inline std::string WildcardSymbol::instance < std::string > ( ) {
	return std::string ( 1, WildcardSymbol::instance < char > ( ) );
}

template < >
inline WildcardSymbol WildcardSymbol::instance < WildcardSymbol > ( ) {
	return WildcardSymbol ( );
}

template < >
inline Symbol WildcardSymbol::instance < Symbol > ( ) {
	return Symbol ( WildcardSymbol ( ) );
}

// TODO make partially specialised when needed by classes or variables, functions can't be partially specialsed
template < >
inline common::ranked_symbol < > WildcardSymbol::instance < common::ranked_symbol < > > ( ) {
	return common::ranked_symbol < > ( WildcardSymbol::instance < DefaultSymbolType > ( ), DefaultRankType ( 0 ) );
}

template < >
inline RankedSymbol < > WildcardSymbol::instance < RankedSymbol < > > ( ) {
	return RankedSymbol < > ( WildcardSymbol::instance < common::ranked_symbol < > > ( ) );
}

} /* namespace alphabet */

#endif /* SUBTREE_WILDCARD_SYMBOL_H_ */
