/*
 * algorithm.hpp
 *
 * Created on: May 28, 2015
 * Author: Jan Travnicek
 */

#ifndef __ALGORITHM_HPP_
#define __ALGORITHM_HPP_

#include <algorithm>
#include <functional>

namespace ext {

template<class InputIt1, class InputIt2, class Compare>
bool excludes(InputIt1 first1, InputIt1 last1, InputIt2 first2, InputIt2 last2, Compare comp) {
	while (first2 != last2 && first1 != last1) {
		if (comp(*first2, *first1)) {
			++first2;
		} else if (comp(*first1, *first2)) {
			++first1;
		} else {
			return false;
		}
	}
	return true;
}

template<class InputIt1, class InputIt2>
bool excludes(InputIt1 first1, InputIt1 last1, InputIt2 first2, InputIt2 last2) {
	return excludes(first1, last1, first2, last2, std::less<decltype(*first1)>());
}

template < class ResType, class InType, typename ... Ts, template < typename ... > class ContainerType, class Callback >
ContainerType < ResType > transform(const ContainerType < InType, Ts ... > & in, Callback transform ) {
	ContainerType<ResType> res;
	std::transform ( in.begin ( ), in.end ( ), std::inserter ( res, res.begin ( ) ), transform );
	return res;
}

template<class InputIt, class Element>
bool contains(InputIt first, InputIt last, const Element& elem) {
	return find(first, last, elem) != last;
}

template<class InputIt, class Element>
bool binary_contains(InputIt first, InputIt last, const Element& elem) {
	return binary_search(first, last, elem) != last;
}

// TODO repace in c++17 with fold expressions
template < class ... Ts >
inline bool andAll ( Ts ... );

template < >
inline bool andAll ( ) {
	return true;
}

template < class T, class ... Ts >
inline bool andAll ( T value, Ts ... other ) {
	return value && andAll ( other ... );
}

// TODO repace in c++17 with fold expressions
template < class ... Ts >
inline bool orAll ( Ts ... );

template < >
inline bool orAll ( ) {
	return false;
}

template < class T, class ... Ts >
inline bool orAll ( T value, Ts ... other ) {
	return value || orAll ( other ... );
}

template < class Iterator, class Value >
std::pair < Iterator, Iterator > __find_range ( Iterator begin, Iterator end, const Value & open, const Value & close ) {
	for ( ; begin != end && * begin != open && * begin != close ; ++ begin );

	if ( begin == end || * begin == close )
		return std::make_pair ( begin, begin );

	Iterator openPos = begin;

	for ( ; ; ) {
		std::pair < Iterator, Iterator > subRange = __find_range ( begin + 1, end, open, close );
		if ( subRange.first == subRange.second )
			break;
		begin = subRange.second;
	}

	if ( begin == end )
		return std::make_pair ( begin, begin );

	++ begin;

	for ( ; begin != end && * begin != close ; ++ begin );

	if ( begin == end )
		return std::make_pair ( begin, begin );

	Iterator closePos = begin;

	return std::make_pair ( openPos, closePos );
}

template < class Iterator, class Value >
std::pair < Iterator, Iterator > find_range ( Iterator begin, Iterator end, const Value & open, const Value & close ) {
	std::pair < Iterator, Iterator > res = __find_range ( begin, end, open, close );

	if ( res.first == res.second )
		return res;

	++ res.second;

	return res;
}

template < typename T >
const T & max ( const T & a ) {
	return a;
}

template < typename T, typename ... Args >
const T & max ( const T & a, const T & b, const Args & ... args ) {
	return max ( b < a ? a : b, args ... );
}

template < typename T >
const T & min ( const T & a) {
	return a;
}

template < typename T, typename ... Args >
const T & min ( const T & a, const T & b, const Args & ... args) {
	return min ( b > a ? a : b, args ... );
}

template < typename _ForwardIterator, typename _BinaryPredicate >
_ForwardIterator __adjacent_find ( _ForwardIterator __first, _ForwardIterator __last, _BinaryPredicate __binary_pred ) {
	if (__first == __last)
		return __last;
	_ForwardIterator __next = __first;
	while (++__next != __last) {
		if ( __binary_pred( * __first, * __next ) )
			return __first;
		__first = __next;
	}
	return __last;
}

template < typename _ForwardIterator, typename _BinaryPredicate >
_ForwardIterator __unique ( _ForwardIterator __first, _ForwardIterator __last, _BinaryPredicate __binary_pred ) {
	// Skip the beginning, if already unique.
	__first = ext::__adjacent_find ( __first, __last, __binary_pred );
	if ( __first == __last )
		return __last;

	// Do the real copy work.
	_ForwardIterator __dest = __first;
	++ __first;
	while ( ++__first != __last )
		if ( ! __binary_pred ( * __dest, * __first ) )
			std::swap ( * ++__dest, * __first );
	return ++ __dest;
}

/**
 *  @brief Shuffles values in a sequece so that consecutive duplicate values are pushed to the front and others to the back.
 *  @ingroup mutating_algorithms
 *  @param  __first  A forward iterator.
 *  @param  __last   A forward iterator.
 *  @return  An iterator designating the end of the resulting sequence.
 *
 *  All but the first element from each group of consecutive
 *  values that compare equal are pushed to the end of the sequence.
 *  unique() is stable, so the relative order of elements that are
 *  not pushed to the end is unchanged.
 *  Elements between the end of the resulting sequence and @p __last
 *  are still present, but their order is unspecified.
*/
template < typename _ForwardIterator >
inline _ForwardIterator unique ( _ForwardIterator __first, _ForwardIterator __last ) {
	return ext::__unique(__first, __last, [] ( const auto & a, const auto & b ) { return a == b; } );
}

/**
 *  @brief Shuffles values in a sequece so that consecutive duplicate values are pushed to the front and others to the back.
 *  @ingroup mutating_algorithms
 *  @param  __first  A forward iterator.
 *  @param  __last   A forward iterator.
 *  @param  __binary_pred  A binary predicate.
 *  @return  An iterator designating the end of the resulting sequence.
 *
 *  All but the first element from each group of consecutive
 *  values for which @p __binary_pred returns true are pushed to the end of the sequence.
 *  unique() is stable, so the relative order of elements that are
 *  not pushed to the end is unchanged.
 *  Elements between the end of the resulting sequence and @p __last
 *  are still present, but their order is unspecified.
*/
template < typename _ForwardIterator, typename _BinaryPredicate >
inline _ForwardIterator unique ( _ForwardIterator __first, _ForwardIterator __last, _BinaryPredicate __binary_pred ) {
	return ext::__unique(__first, __last, __binary_pred );
}

} /* namespace ext */

#endif /* __ALGORITHM_HPP_ */
