/*
 * range.hpp as proposed http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2012/n3350.html by Jeffrey Yasskin <jyasskin@google.com>
 *
 * Created on: May 28, 2015
 * Author: Jan Travnicek
 */

#ifndef __RANGE_HPP_
#define __RANGE_HPP_

#include <utility>
#include <iterator>

namespace ext {

template<typename Iterator>
class range {
	Iterator m_begin;
	Iterator m_end;
public:
	// types
	typedef typename std::iterator_traits< Iterator >::iterator_category iterator_category;
	typedef typename std::iterator_traits< Iterator >::value_type value_type;
	typedef typename std::iterator_traits< Iterator >::difference_type difference_type;
	typedef typename std::iterator_traits< Iterator >::reference reference;
	typedef typename std::iterator_traits< Iterator >::pointer pointer;

	// constructors
	range() {
	}

	constexpr range(Iterator begin, Iterator end) : m_begin ( begin ), m_end ( end ) {
	}

	// iterator access
	constexpr Iterator begin() const {
		return m_begin;
	}

	constexpr Iterator end() const {
		return m_end;
	}

	// element access
	constexpr reference front() const {
		return * m_begin;
	}

	constexpr reference back() const {
		return * ( m_end - 1 );
	}

	constexpr reference operator[](difference_type index) const {
		return m_begin [ index ];
	}

	// size
	constexpr bool empty() const {
		return m_begin == m_end;
	}

	constexpr size_t size() const {
		return std::distance ( m_begin, m_end );
	}

	// traversal from the beginning of the range
	void pop_front() {
		++ m_begin;
	}

	void pop_front(difference_type n) {
		m_begin = m_begin + n;
	}

	// traversal from the end of the range
	void pop_back() {
		-- m_end;
	}

	void pop_back(difference_type n) {
		m_end = m_end - n;
	}

	// creating derived ranges
	std::pair< range, range > split(difference_type index) const {
		return std::make_pair ( slice ( 0, index ), slice ( index, 0 ) );
	}

	range slice(difference_type start, difference_type stop) const {
		return range ( ( start >= 0 ? m_begin : m_end ) + start, ( stop > 0 ? m_begin : m_end ) + stop );
	}

	range slice(difference_type start) const {
		return slice ( start, 0 );
	}
};

template < typename Iter >
range < Iter > make_range ( Iter begin, Iter end ) {
	return range < Iter > ( begin, end );
}

template < typename Container >
auto make_range ( Container & container ) {
	return make_range ( container.begin ( ), container.end ( ) );
}

template < typename Container >
auto make_range ( const Container & container ) {
	return make_range ( container.begin ( ), container.end ( ) );
}

} /* namespace ext */

#endif /* __RANGE_HPP_ */

