/*
 * tuple.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef __TUPLE_HPP_
#define __TUPLE_HPP_

#include <tuple>
#include <ostream>
#include <sstream>
#include <string>

#include "compare.hpp"

namespace ext {

template < typename ... Ts >
class tuple : public std::tuple < Ts ... > {
public:
#ifdef __clang__
	using std::tuple< Ts ... >::tuple;
	using std::tuple< Ts ... >::operator =;
#else
	tuple ( ) noexcept : std::tuple < Ts ... > ( ) {
	}

	tuple ( const tuple & other ) = default;

	tuple ( tuple && other ) = default;

	using std::tuple < Ts ... >::tuple;

	tuple & operator = ( tuple && other ) = default;

	tuple & operator = ( const tuple & other ) = default;

	using std::tuple < Ts ... >::operator =;
#endif
};

template<typename... _Elements>
constexpr tuple < typename std::__decay_and_strip < _Elements >::__type... > make_tuple ( _Elements && ... __args ) {
	typedef tuple < typename std::__decay_and_strip < _Elements >::__type... > __result_type;
	return __result_type ( std::forward < _Elements > ( __args ) ... );
}

template<typename ... _Elements >
constexpr tuple < _Elements & ... > tie ( _Elements & ... __args ) noexcept {
	return tuple < _Elements & ... > ( __args ... );
}

template < class ... Types >
class tuple_size;

template < class ... Types >
class tuple_size < ext::tuple < Types ... > > : public std::integral_constant < std::size_t, sizeof...( Types ) > { };

template < class T >
class tuple_size < T & > : public std::integral_constant < std::size_t, ext::tuple_size < T >::value > { };

template < class T >
class tuple_size < T && > : public std::integral_constant < std::size_t, ext::tuple_size < T >::value > { };

template < class T >
class tuple_size < const T > : public std::integral_constant < std::size_t, ext::tuple_size < T >::value > { };

template < class T >
class tuple_size< volatile T > : public std::integral_constant < std::size_t, ext::tuple_size < T >::value > { };

template< class T >
class tuple_size < const volatile T > : public std::integral_constant < std::size_t, tuple_size < T >::value > { };


template < std::size_t I, class T >
struct tuple_element;

template < std::size_t I, class... Types >
struct tuple_element < I, ext::tuple < Types... > > {
	typedef typename std::tuple_element < I, std::tuple < Types ... > >::type type;
};

template < std::size_t I, class T >
struct tuple_element < I, const T > {
	typedef typename std::add_const < typename ext::tuple_element < I, T >::type >::type type;
};

template < std::size_t I, class T >
struct tuple_element< I, volatile T > {
	typedef typename std::add_volatile < typename ext::tuple_element < I, T >::type >::type type;
};

template < std::size_t I, class T >
struct tuple_element < I, const volatile T > {
	typedef typename std::add_cv < typename ext::tuple_element < I, T >::type >::type type;
};

template < class Result, unsigned I, class Tuple, class Callable >
struct call_on_nth_helper;

template < class Result, unsigned I, class Callable, class ... Types >
struct call_on_nth_helper < Result, I, ext::tuple < Types ... >, Callable > {
	template < class Tuple >
	static Result call_on_nth_fn ( Tuple && t, unsigned index, Callable callback ) {
		if ( index == 0 )
			return callback ( std::get < ext::tuple_size < Tuple >::value - I > ( std::forward < Tuple && > ( t ) ) );
		else
			return call_on_nth_helper < Result, I - 1, ext::tuple < Types ... >, Callable >::call_on_nth_fn ( std::forward < Tuple && > ( t ), index - 1, callback );
	}
};

template < class Result, class Callable, class ... Types >
struct call_on_nth_helper < Result, 0, ext::tuple < Types ... >, Callable > {
	template < class Tuple >
	static Result call_on_nth_fn ( Tuple &&, unsigned, Callable ) {
		throw std::out_of_range ( "Not enough elements in tuple." );
	}
};

template < class Result, class Tuple, class Callable >
Result call_on_nth ( Tuple && t, unsigned index, Callable callback ) {
	return call_on_nth_helper < Result, ext::tuple_size < Tuple >::value, typename std::decay < Tuple >::type, Callable >::call_on_nth_fn ( std::forward < Tuple && > ( t ), index, callback );
}



template < unsigned I, class Tuple, class Callable >
struct foreach_helper;

template < unsigned I, class Callable, class ... Types >
struct foreach_helper < I, ext::tuple < Types ... >, Callable > {
	template < class Tuple >
	static void foreach_fn ( Tuple && t, Callable callback ) {
		foreach_helper < I - 1, ext::tuple < Types ... >, Callable >::foreach_fn ( std::forward < Tuple && > ( t ), callback );
		callback ( std::get < I - 1 > ( t ) );
	}
};

template < class Callable, class ... Types >
struct foreach_helper < 0, ext::tuple < Types ... >, Callable > {
	template < class Tuple >
	static void foreach_fn ( Tuple &&, Callable ) {
	}
};

template < class Tuple, class Callable >
void foreach ( Tuple && t, Callable callback ) {
	return foreach_helper < ext::tuple_size < Tuple >::value, typename std::decay < Tuple >::type, Callable >::foreach_fn ( std::forward < Tuple && > ( t ), callback );
}



template < class Tuple, class Callable >
bool all_of ( const Tuple & t, Callable callback ) {
	bool res = true;

	auto aggregateCallback = [ & ] ( const auto & arg0 ) {
		res &= callback ( arg0 );
	};

	ext::foreach ( t, aggregateCallback  );
	return res;
}

template< class... Ts>
std::ostream& operator<<(std::ostream& out, const ext::tuple<Ts...>& tuple) {
	out << "(";

	bool first = true;
	auto outCallback = [ & ] ( const auto & arg0 ) {
		if ( ! first ) {
			out << ", ";
		} else {
			first = false;
		}
		out << arg0;
	};

	ext::foreach ( tuple, outCallback  );
	out << ")";
	return out;
}

template < int I, typename Tuple >
struct compareTupleHelper;

template < int I, typename Tuple >
struct compareTupleHelper {
	static int compHelp ( const Tuple & t1, const Tuple & t2 ) {
		int res = compareTupleHelper < I - 1, Tuple >::compHelp ( t1, t2 );

		if ( res != 0 ) return res;

		static compare < typename std::decay < typename ext::tuple_element < I, Tuple >::type >::type > comp;
		return comp ( std::get < I > ( t1 ), std::get < I > ( t2 ) );
	}

};

template < class Tuple >
struct compareTupleHelper < 0, Tuple > {
	static int compHelp ( const Tuple & t1, const Tuple & t2 ) {
		static compare < typename std::decay < typename ext::tuple_element < 0, Tuple >::type >::type > comp;
		return comp ( std::get < 0 > ( t1 ), std::get < 0 > ( t2 ) );
	}

};

template < typename ... Ts >
struct compare < ext::tuple < Ts ... > > {
	int operator ()( const ext::tuple < Ts ... > & first, const ext::tuple < Ts ... > & second ) const {
		return compareTupleHelper < ext::tuple_size < ext::tuple < Ts ... > >::value - 1, ext::tuple < Ts ... > >::compHelp ( first, second );
	}

};

template < class ... Ts >
std::string to_string ( const ext::tuple < Ts ... > & value ) {
	std::stringstream ss;
	ss << value;
	return ss.str();
}

template < class Type, int size, class ... Types >
struct TupleBuilder;

template < class Type, class ... Types >
struct TupleBuilder < Type, 0, Types ... > {
	typedef ext::tuple < Types ... > type;
};

template < class Type, int n, class ... Types >
struct TupleBuilder : public TupleBuilder < Type, n - 1, Type, Types ... > {
};

} /* namespace ext */

#endif /* __TUPLE_HPP_ */
